/*
 *                   jfjji,.           H A M M E R H E A D  S H A R K W A R E
 *                  ,#######EL:                  http://www.hhsw.de
 *                   .,j######Kj       
 *                    ;W########G:     This  code  is  free software; you can
 *                   j#####G ,D#L:     redistribute it and/or modify it under
 *             .,;;,G#####D.           the  terms  of  the GNU General Public
 *           jE##########G             License   as  published  by  the  Free
 *             .,W######D.             Software  Foundation; either version 2
 *             :L#####K##G             of  the  License,  or (at your option)
 *           .f#####G;.t##:            any later version.
 *         ;L###ELi.    ,D,            
 *       iEEGWj                        This  code  is distributed in the hope
 *      tW;                            that  it  will  be useful, but without
 *     jf                              any warranty; without even the implied
 *    .Kt                              warranty of merchantability or fitness
 *    i:                               for  a particular purpose. See the GNU
 *    ;                                General   Public   License   for  more
 *    .                                details (http://www.gnu.org).
 */
#include "Render.hh"

using namespace bbdock;

/**
 * Return a Render object for the demanded color depth
 *
 * @param bitsperpixel - bits per pixel (color depth)
 * @param width - width of the image in pixels
 * @param height - height of the image in pixels
 */
Render *Render::getInstance( int bitsperpixel, int width, int height )
{
	switch( bitsperpixel )
	{
		default:
			throw "Pixel depth not (yet) supported";
		case 16:
			return (Render *) new Render16( width, height );
		case 24:
			return (Render *) new Render24( width, height );
		case 32:
			return (Render *) new Render32( width, height );
	}
}

/**
 * Initialize internal parameters
 *
 * @param bitsperpixel - bits per pixel (color depth)
 * @param width - width of the image in pixels
 * @param height - height of the image in pixels
 */
void Render::initialize( int bitsperpixel, int width, int height )
{
	this->width = width;
	this->height = height;

	// ceil to full byte if bits per pixel is fewer than a byte
	// or divide by 8 if there are at last 8 bits per pixel
	if( bitsperpixel < 8 )
		bytesperline = ((width*bitsperpixel)+7) & ~7;
	else
		bytesperline = (width*bitsperpixel)>>3;

	// pad every image line to a multiple of 4 bytes
	padding = bytesperline;
	bytesperline = (bytesperline+3) & ~3;
	padding = bytesperline-padding;

	size = bytesperline*height;
	pixels = width*height;
}

/**
 * Initialize object
 *
 * @param width - width of the image in pixels
 * @param height - height of the image in pixels
 */
Render16::Render16( int width, int height )
{
	initialize( 16, width, height );
}

/**
 * Copy a rgba image into a 16 bit pixmap
 *
 * @param dest - destination pixels
 * @param src - source pixels (in rgba format !)
 */
void Render16::opaque( unsigned char *dest, unsigned char *src )
{
	for( int y = height; y--; dest += padding )
		for( int x = width; x--; )
			switch( *((int *) src) & 0xff000000 )
			{
				case 0:
					src += 4;
					dest += 2;
					break;
				case 0xff000000:
					{
						register int blue = *(src++);
						register int green = *(src++);
						register int red = *(src++);

						*((short *) dest) = 
							((short) ((blue&0xf8)>>3)) |
							((short) ((green&0xf8)<<3)) |
							((short) ((red&0xf8)<<8));

						src++;
						dest += 2;
					}
					break;
				default:
					{
						double	diff;
						double	mod = (double) 255/(double) (*(src+3));

						short	pixel = *((short *) dest);

						register int blue = (pixel<<3)&0xf8;
						register int green = (pixel>>3)&0xf8;
						register int red = (pixel>>8)&0xf8;

						diff = *(src++);
						diff -= blue;
						diff /= mod;
						blue += (int) diff;

						diff = *(src++);
						diff -= green;
						diff /= mod;
						green += (int) diff;

						diff = *(src++);
						diff -= red;
						diff /= mod;
						red += (int) diff;

						*((short *) dest) = 
							((short) ((blue&0xf8)>>3)) |
							((short) ((green&0xf8)<<3)) |
							((short) ((red&0xf8)<<8));

						dest += 2;
						src++;
					}
					break;
			}
}

/**
 * Copy a rgba image greyed into a 16 bit pixmap respecting transparency
 *
 * @param dest - destination pixels
 * @param src - source pixels (in rgba format !)
 * @param transparency - global transparency
 */
void Render16::ghosted( unsigned char *dest, unsigned char *src, 
	unsigned char transparency )
{
	double	alphamax = ((double) 255*((double) 255/(double) transparency));

	for( int y = height; y--; dest += padding )
		for( int x = width; x--; )
			if( !(*((int *) src) & 0xff000000) )
			{
				src += 4;
				dest += 2;
			}
			else
			{
				double	diff;
				double	mod = alphamax/(double) (*(src+3));

				// it must be done that way, if you write it in just
				// line, things get messed up terribly by the compiler 
				// (at least to version 3.3)
				register int red = *(src++);
				register int green = *(src++);
				register int blue = *(src++);
				register int grey = (red+green+blue)/3;

				short	pixel = *((short *) dest);

				blue = (pixel<<3)&0xf8;
				green = (pixel>>3)&0xf8;
				red = (pixel>>8)&0xf8;

				diff = grey;
				diff -= blue;
				diff /= mod;
				blue += (int) diff;

				diff = grey;
				diff -= green;
				diff /= mod;
				green += (int) diff;

				diff = grey;
				diff -= red;
				diff /= mod;
				red += (int) diff;

				*((short *) dest) = 
					((short) ((blue&0xf8)>>3)) |
					((short) ((green&0xf8)<<3)) |
					((short) ((red&0xf8)<<8));

				dest += 2;
				src++;
			}
}

/**
 * Copy a rgba image into a 16 bit surface and mark it as active
 *
 * @param dest - destination pixels
 * @param src - source pixels (in rgba format !)
 */
void Render16::active( unsigned char *dest, unsigned char *src )
{
	opaque( dest, src );

	// draw a litte arrow in upper left corner
	for( int i = 0; i++ < 5; dest += bytesperline )
	{
		short	*p = (short *) dest;

		for( int x = i; --x; )
			*(p++) ^= 0xfffe;
	}

	for( int i = 5; --i; dest += bytesperline )
	{
		short	*p = (short *) dest;

		for( int x = i; --x; )
			*(p++) ^= 0xfffe;
	}
}

/**
 * Initialize object
 *
 * @param width - width of the image in pixels
 * @param height - height of the image in pixels
 */
Render24::Render24( int width, int height )
{
	initialize( 24, width, height );
}

/**
 * Copy a rgba image into a 24 bit pixmap
 *
 * @param dest - destination pixels
 * @param src - source pixels (in rgba format !)
 */
void Render24::opaque( unsigned char *dest, unsigned char *src )
{
	for( int y = height; y--; dest += padding )
		for( int x = width; x--; )
			switch( *((int *) src) & 0xff000000 )
			{
				case 0:
					src += 4;
					dest += 3;
					break;
				case 0xff000000:
					*(dest++) = *(src++);
					*(dest++) = *(src++);
					*(dest++) = *(src++);
					src++;
					break;
				default:
					{
						double	diff;
						double	mod = (double) 255/(double) (*(src+3));

						diff = *(src++);
						diff -= *dest;
						diff /= mod;
						*(dest++) += (int) diff;

						diff = *(src++);
						diff -= *dest;
						diff /= mod;
						*(dest++) += (int) diff;

						diff = *(src++);
						diff -= *dest;
						diff /= mod;
						*(dest++) += (int) diff;

						src++;
					}
					break;
			}
}

/**
 * Copy a rgba image greyed into a 24 bit pixmap respecting transparency
 *
 * @param dest - destination pixels
 * @param src - source pixels (in rgba format !)
 * @param transparency - global transparency
 */
void Render24::ghosted( unsigned char *dest, unsigned char *src, 
	unsigned char transparency )
{
	double	alphamax = ((double) 255*((double) 255/(double) transparency));

	for( int y = height; y--; dest += padding )
		for( int x = width; x--; )
			if( !(*((int *) src) & 0xff000000) )
			{
				src += 4;
				dest += 3;
			}
			else
			{
				double	diff;
				double	mod = alphamax/(double) (*(src+3));

				// it must be done that brain-dead, if you write it in just
				// line, things get messed up terribly by the compiler 
				// (at least to version 3.3)
				register int red = *(src++);
				register int green = *(src++);
				register int blue = *(src++);
				register int grey = (red+green+blue)/3;

				diff = grey;
				diff -= *dest;
				diff /= mod;
				*(dest++) += (int) diff;

				diff = grey;
				diff -= *dest;
				diff /= mod;
				*(dest++) += (int) diff;

				diff = grey;
				diff -= *dest;
				diff /= mod;
				*(dest++) += (int) diff;

				src++;
			}
}

/**
 * Copy a rgba image into a 24 bit surface and mark it as active
 *
 * @param dest - destination pixels
 * @param src - source pixels (in rgba format !)
 */
void Render24::active( unsigned char *dest, unsigned char *src )
{
	opaque( dest, src );

	// draw a litte arrow in upper left corner
	for( int i = 0; i++ < 5; dest += bytesperline )
	{
		unsigned char	*p = dest;

		for( int x = i; --x; )
		{
			*(p++) ^= 0xff;
			*(p++) ^= 0xff;
			*(p++) ^= 0xff;
		}
	}

	for( int i = 5; --i; dest += bytesperline )
	{
		unsigned char	*p = dest;

		for( int x = i; --x; )
		{
			*(p++) ^= 0xff;
			*(p++) ^= 0xff;
			*(p++) ^= 0xff;
		}
	}
}

/**
 * Initialize object
 *
 * @param width - width of the image in pixels
 * @param height - height of the image in pixels
 */
Render32::Render32( int width, int height )
{
	initialize( 32, width, height );
}

/**
 * Copy a rgba image into a 32 bit pixmap
 *
 * @param dest - destination pixels
 * @param src - source pixels (in rgba format !)
 */
void Render32::opaque( unsigned char *dest, unsigned char *src )
{
	for( int p = pixels; p--; )
		switch( *((int *) src) & 0xff000000 )
		{
			case 0:
				src += 4;
				dest += 4;
				break;
			case 0xff000000:
				*((int *) dest) = *((int *) src);
				src += 4;
				dest += 4;
				break;
			default:
				{
					double	diff;
					double	mod = (double) 255/(double) (*(src+3));

					diff = *(src++);
					diff -= *dest;
					diff /= mod;
					*(dest++) += (int) diff;

					diff = *(src++);
					diff -= *dest;
					diff /= mod;
					*(dest++) += (int) diff;

					diff = *(src++);
					diff -= *dest;
					diff /= mod;
					*(dest++) += (int) diff;

					src++;
					dest++;
				}
				break;
		}
}

/**
 * Copy a rgba image greyed into a 32 bit pixmap respecting transparency
 *
 * @param dest - destination pixels
 * @param src - source pixels (in rgba format !)
 * @param transparency - global transparency
 */
void Render32::ghosted( unsigned char *dest, unsigned char *src, 
	unsigned char transparency )
{
	double	alphamax = ((double) 255*((double) 255/(double) transparency));

	for( int p = pixels; p--; )
		if( !(*((int *) src) & 0xff000000) )
		{
			src += 4;
			dest += 4;
		}
		else
		{
			double	diff;
			double	mod = alphamax/(double) (*(src+3));

			// it must be done that brain-dead, if you write it in just
			// line, things get messed up terribly by the compiler 
			// (at least to version 3.3)
			register int red = *(src++);
			register int green = *(src++);
			register int blue = *(src++);
			register int grey = (red+green+blue)/3;

			diff = grey;
			diff -= *dest;
			diff /= mod;
			*(dest++) += (int) diff;

			diff = grey;
			diff -= *dest;
			diff /= mod;
			*(dest++) += (int) diff;

			diff = grey;
			diff -= *dest;
			diff /= mod;
			*(dest++) += (int) diff;

			src++;
			dest++;
		}
}

/**
 * Copy a rgba image into a 32 bit surface and mark it as active
 *
 * @param dest - destination pixels
 * @param src - source pixels (in rgba format !)
 */
void Render32::active( unsigned char *dest, unsigned char *src )
{
	opaque( dest, src );

	// draw a litte arrow in upper left corner
	for( int i = 0; i++ < 5; dest += bytesperline )
	{
		int	*p = (int *) dest;

		for( int x = i; --x; )
			*(p++) ^= 0xffffff;
	}

	for( int i = 5; --i; dest += bytesperline )
	{
		int	*p = (int *) dest;

		for( int x = i; --x; )
			*(p++) ^= 0xffffff;
	}
}
